
RELEASE_VERSION='0.3.568835525'
BUILD_DATE='Tue Jun 21 09:38:28 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-opyxurxq-project-28359933-concurrent-0sx4g2 5.4.181-99.354.amzn2.x86_64 #1 SMP Wed Mar 2 18:50:46 UTC 2022 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2617785041'
GIT_COMMIT_SHA='4c1a5801fc326011d14da2314e04da054cebf8e1'
GIT_BRANCH='main'
